package semmiedev.disc_jockey_revive;

import net.fabricmc.fabric.api.client.event.lifecycle.v1.ClientTickEvents;
import net.minecraft.block.Block;
import net.minecraft.block.BlockState;
import net.minecraft.block.Blocks;
import net.minecraft.block.enums.NoteBlockInstrument;
import net.minecraft.client.MinecraftClient;
import net.minecraft.client.gui.hud.ChatHud;
import net.minecraft.client.network.ClientPlayerEntity;
import net.minecraft.client.network.PlayerListEntry;
import net.minecraft.client.world.ClientWorld;
import net.minecraft.network.packet.c2s.play.PlayerActionC2SPacket;
import net.minecraft.network.packet.c2s.play.PlayerMoveC2SPacket;
import net.minecraft.state.property.Properties;
import net.minecraft.text.Text;
import net.minecraft.util.Formatting;
import net.minecraft.util.Hand;
import net.minecraft.util.Pair;
import net.minecraft.util.hit.BlockHitResult;
import net.minecraft.util.math.*;
import net.minecraft.world.GameMode;
import org.apache.commons.lang3.NotImplementedException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

public class SongPlayer implements ClientTickEvents.StartWorldTick {
    private static boolean warned;
    public boolean running;
    public Song song;

    public int index;
    public double tick;
    public HashMap<NoteBlockInstrument, HashMap<Byte, BlockPos>> noteBlocks = null;
    public boolean tuned;
    private long lastPlaybackTickAt = -1L;

    // Used to check and enforce packet rate limits to not get kicked
    private long last100MsSpanAt = -1L;
    private int last100MsSpanEstimatedPackets = 0;
    // At how many packets/100ms should the player just reduce / stop sending packets for a while
    final private int last100MsReducePacketsAfter = 300 / 10, last100MsStopPacketsAfter = 450 / 10;
    // If higher than current millis, don't send any packets of this kind (temp disable)
    private long reducePacketsUntil = -1L, stopPacketsUntil = -1L;

    // Use to limit swings and look to only each tick. More will not be visually visible anyway due to interpolation
    private long lastLookSentAt = -1L, lastSwingSentAt = -1L;

    // The thread executing the tickPlayback method
    private Thread playbackThread = null;
    public long playbackLoopDelay = 5;
    // Just for external debugging purposes
    public HashMap<Block, Integer> missingInstrumentBlocks = new HashMap<>();
    public float speed = 1.0f; // Toy

    private long lastInteractAt = -1;
    private float availableInteracts = 8;
    private int tuneInitialUntunedBlocks = -1;
    private HashMap<BlockPos, Pair<Integer, Long>> notePredictions = new HashMap<>();
    public boolean didSongReachEnd = false;
    public boolean loopSong = false;
    private long pausePlaybackUntil = -1L;

    private boolean manualTuningRequested = false;


    public SongPlayer() {
        Main.TICK_LISTENERS.add(this);
    }

    public @NotNull HashMap<NoteBlockInstrument, @Nullable NoteBlockInstrument> instrumentMap = new HashMap<>(); // Toy
    public synchronized void startPlaybackThread() {
        if(Main.config.disableAsyncPlayback) {
            playbackThread = null;
            return;
        }

        this.playbackThread = new Thread(() -> {
            Thread ownThread = this.playbackThread;
            while(ownThread == this.playbackThread) {
                try {
                    // Accuracy doesn't really matter at this precision imo
                    Thread.sleep(playbackLoopDelay);
                }catch (Exception ex) {
                    ex.printStackTrace();
                }
                MinecraftClient.getInstance().executeSync(this::tickPlayback);
            }
        });
        this.playbackThread.start();
    }

    public enum PlayMode {
        SINGLE_LOOP, // 单曲循环
        LIST_LOOP,   // 列表循环
        RANDOM,      // 随机播放
        STOP_AFTER   // 播完就停
    }

    public PlayMode playMode = PlayMode.STOP_AFTER;
    private boolean isRandomPlaying = false;
    private int randomIndex = -1;

    public synchronized void stopPlaybackThread() {
        this.playbackThread = null; // Should stop on its own then
    }

    public synchronized void start(Song song) {
        if (!Main.config.hideWarning && !warned) {
            MinecraftClient.getInstance().inGameHud.getChatHud().addMessage(Text.translatable("disc_jockey_revive.warning").formatted(Formatting.BOLD, Formatting.RED));
            warned = true;
            return;
        }
        if (running) stop();
        this.song = song;
        //Main.LOGGER.info("Song length: " + song.length + " and tempo " + song.tempo);
        //Main.TICK_LISTENERS.add(this);
        this.loopSong = playMode == PlayMode.SINGLE_LOOP;
        if(this.playbackThread == null) startPlaybackThread();
        running = true;
        index = 0;
        tick = 0;
        startTuning();

        lastPlaybackTickAt = System.currentTimeMillis();
        didSongReachEnd = false;
        isRandomPlaying = playMode == PlayMode.RANDOM;
        if (isRandomPlaying) {
            randomIndex = getRandomSongIndex();
        }
    }

    private int getRandomSongIndex() {
        if (SongLoader.currentFolder == null) {
            return (int) (Math.random() * SongLoader.SONGS.size());
        } else {
            return (int) (Math.random() * SongLoader.currentFolder.songs.size());
        }
    }

    public synchronized void stop() {
        //MinecraftClient.getInstance().send(() -> Main.TICK_LISTENERS.remove(this));
        stopPlaybackThread();
        running = false;
        index = 0;
        tick = 0;
        noteBlocks = null;
        notePredictions.clear();
        tuned = false;
        tuneInitialUntunedBlocks = -1;
        manualTuningRequested = false;

        lastPlaybackTickAt = -1L;
        // last100MsSpanAt = -1L;
        // last100MsSpanEstimatedPackets = 0;
        // reducePacketsUntil = -1L;
        // stopPacketsUntil = -1L;
        // lastLookSentAt = -1L;
        // lastSwingSentAt = -1L;
        didSongReachEnd = false; // Change after running stop() if actually ended cleanly
    }

    public synchronized void startTuning() {
        if (tuned && noteBlocks != null) {
            MinecraftClient.getInstance().inGameHud.getChatHud().addMessage(Text.translatable(Main.MOD_ID+".player.retuning").formatted(Formatting.YELLOW));
        } else {
            MinecraftClient.getInstance().inGameHud.getChatHud().addMessage(Text.translatable(Main.MOD_ID+".player.tuning_started").formatted(Formatting.YELLOW));
        }
        noteBlocks = null;
        notePredictions.clear();
        tuned = false;
        tuneInitialUntunedBlocks = -1;
        manualTuningRequested = true;
    }

    public boolean isTuningEnabled() {
        return running || manualTuningRequested;
    }


    public boolean playNoteBlock(Note note) {
        MinecraftClient client = MinecraftClient.getInstance();
        ClientWorld world = client.world;
        ClientPlayerEntity player = client.player;
        GameMode gameMode = client.interactionManager == null ? null : client.interactionManager.getCurrentGameMode();

        if (world == null || player == null || gameMode == null || !gameMode.isSurvivalLike()) {
            return false;
        }

        if (noteBlocks == null || !tuned) {
            return false;
        }

        @Nullable BlockPos blockPos = noteBlocks.get(note.instrument()).get(note.note());
        if(blockPos == null) {
            return false;
        }

        if (!canInteractWith(player, blockPos)) {
            return false;
        }

        return sendNotePacket(blockPos, note);
    }


    private boolean sendNotePacket(BlockPos blockPos, Note note) {
        MinecraftClient client = MinecraftClient.getInstance();
        ClientPlayerEntity player = client.player;
        long now = System.currentTimeMillis();

        // Update packet rate tracking
        if(last100MsSpanAt != -1L && now - last100MsSpanAt >= 100) {
            last100MsSpanEstimatedPackets = 0;
            last100MsSpanAt = now;
        }else if (last100MsSpanAt == -1L) {
            last100MsSpanAt = now;
            last100MsSpanEstimatedPackets = 0;
        }

        if (stopPacketsUntil != -1L && stopPacketsUntil >= now) {
            return false;
        }
        if (reducePacketsUntil != -1L && reducePacketsUntil >= now && last100MsSpanEstimatedPackets >= last100MsReducePacketsAfter) {
            return false;
        }

        if(lastInteractAt != -1L) {
            availableInteracts += ((System.currentTimeMillis() - lastInteractAt) / (310.0f / 8.0f));
            availableInteracts = Math.min(8f, Math.max(0f, availableInteracts));
        }else {
            availableInteracts = 8f;
            lastInteractAt = System.currentTimeMillis();
        }

        if (availableInteracts < 1f) {
            return false;
        }


        Vec3d unit = Vec3d.ofCenter(blockPos, 0.5).subtract(player.getEyePos()).normalize();
        boolean packetsSent = false;

        if((lastLookSentAt == -1L || now - lastLookSentAt >= 50) && last100MsSpanEstimatedPackets < last100MsReducePacketsAfter) {
            client.getNetworkHandler().sendPacket(new PlayerMoveC2SPacket.LookAndOnGround(MathHelper.wrapDegrees((float) (MathHelper.atan2(unit.z, unit.x) * 57.2957763671875) - 90.0f), MathHelper.wrapDegrees((float) (-(MathHelper.atan2(unit.y, Math.sqrt(unit.x * unit.x + unit.z * unit.z)) * 57.2957763671875))), true, false));
            last100MsSpanEstimatedPackets++;
            lastLookSentAt = now;
            packetsSent = true;
        } else if (last100MsSpanEstimatedPackets >= last100MsReducePacketsAfter) {
            reducePacketsUntil = Math.max(reducePacketsUntil, now + 500);
        }


        if(last100MsSpanEstimatedPackets < last100MsStopPacketsAfter) {
            client.player.networkHandler.sendPacket(new PlayerActionC2SPacket(PlayerActionC2SPacket.Action.START_DESTROY_BLOCK, blockPos, Direction.UP, 0));
            last100MsSpanEstimatedPackets++;
            client.player.networkHandler.sendPacket(new PlayerActionC2SPacket(PlayerActionC2SPacket.Action.ABORT_DESTROY_BLOCK, blockPos, Direction.UP, 0));
            last100MsSpanEstimatedPackets++;
            lastInteractAt = now;
            availableInteracts -= 1f;
            packetsSent = true;
        } else {
            Main.LOGGER.info("短时间内暂停所有数据包！");
            stopPacketsUntil = Math.max(stopPacketsUntil, now + 250);
            reducePacketsUntil = Math.max(reducePacketsUntil, now + 10000);
        }


        if((lastSwingSentAt == -1L || now - lastSwingSentAt >= 50) && last100MsSpanEstimatedPackets < last100MsReducePacketsAfter) {
            client.executeSync(() -> client.player.swingHand(Hand.MAIN_HAND));
            lastSwingSentAt = now;
            last100MsSpanEstimatedPackets++;
            packetsSent = true;
        } else if (last100MsSpanEstimatedPackets >= last100MsReducePacketsAfter) {
            reducePacketsUntil = Math.max(reducePacketsUntil, now + 500);
        }

        return packetsSent;
    }


    public synchronized void tickPlayback() {
        if (!running || song == null) {
            lastPlaybackTickAt = -1L;
            // last100MsSpanAt = -1L;
            return;
        }

        if (pausePlaybackUntil != -1L) {
            if (System.currentTimeMillis() <= pausePlaybackUntil) {
                return;
            } else {
                tick = 0;
                index = 0;
                pausePlaybackUntil = -1L;
                lastPlaybackTickAt = System.currentTimeMillis();
            }
        }

        long currentNow = System.currentTimeMillis();
        long elapsedMs = lastPlaybackTickAt != -1L ? currentNow - lastPlaybackTickAt : (16);
        tick += song.millisecondsToTicks(elapsedMs) * speed;
        lastPlaybackTickAt = currentNow;


        if(noteBlocks != null && tuned) {
            while (index < song.notes.length) {
                long note = song.notes[index];

                if ((double)(short)note <= tick) {
                    Note currentNote = new Note(Note.INSTRUMENTS[(byte)(note >> Note.INSTRUMENT_SHIFT)], (byte)(note >> Note.NOTE_SHIFT));

                    playNoteBlock(currentNote);

                    index++;
                } else {
                    break;
                }
            }

            if (index >= song.notes.length) {
                stop();
                didSongReachEnd = true;
                if (playMode == PlayMode.SINGLE_LOOP) {
                } else if (playMode == PlayMode.LIST_LOOP || playMode == PlayMode.RANDOM) {
                    playNextSong();
                }
            }
        }
    }

    private void playNextSong() {
        if (SongLoader.currentFolder == null || SongLoader.currentFolder.songs.isEmpty()) return;

        int currentIndex = SongLoader.currentFolder.songs.indexOf(song);
        if (currentIndex == -1) {
            if (!SongLoader.currentFolder.songs.isEmpty()) {
                start(SongLoader.currentFolder.songs.get(0));
            }
            return;
        }

        if (playMode == PlayMode.RANDOM) {
            int newIndex;
            do {
                newIndex = (int) (Math.random() * SongLoader.currentFolder.songs.size());
            } while (newIndex == currentIndex && SongLoader.currentFolder.songs.size() > 1);
            start(SongLoader.currentFolder.songs.get(newIndex));
        } else if (playMode == PlayMode.LIST_LOOP) {
            int nextIndex = (currentIndex + 1) % SongLoader.currentFolder.songs.size();
            start(SongLoader.currentFolder.songs.get(nextIndex));
        }
    }

    public synchronized void setPlayMode(PlayMode mode) {
        this.playMode = mode;
        this.loopSong = mode == PlayMode.SINGLE_LOOP;
    }

    // 我都不知道这是啥时候的，我也不删
    // TODO: 6/2/2022 Play note blocks every song tick, instead of every tick. That way the song will sound better
    //      11/1/2023 Playback now done in separate thread. Not ideal but better especially when FPS are low.
    @Override
    public void onStartTick(ClientWorld world) {
        MinecraftClient client = MinecraftClient.getInstance();
        if(world == null || client.world == null || client.player == null) return;
        long now = System.currentTimeMillis();
        if(last100MsSpanAt != -1L && now - last100MsSpanAt >= 100) {
            last100MsSpanEstimatedPackets = 0;
            last100MsSpanAt = now;
        }else if (last100MsSpanAt == -1L) {
            last100MsSpanAt = now;
            last100MsSpanEstimatedPackets = 0;
        }

        if(lastInteractAt != -1L) {
            availableInteracts += ((System.currentTimeMillis() - lastInteractAt) / (310.0f / 8.0f));
            availableInteracts = Math.min(8f, Math.max(0f, availableInteracts));
        }else {
            availableInteracts = 8f;
            lastInteractAt = System.currentTimeMillis();
        }


        // Clear outdated note predictions
        ArrayList<BlockPos> outdatedPredictions = new ArrayList<>();
        for(Map.Entry<BlockPos, Pair<Integer, Long>> entry : notePredictions.entrySet()) {
            if(entry.getValue().getRight() < System.currentTimeMillis())
                outdatedPredictions.add(entry.getKey());
        }
        for(BlockPos outdatedPrediction : outdatedPredictions) notePredictions.remove(outdatedPrediction);

        if ((noteBlocks == null || !tuned) && (running || manualTuningRequested)) {
            ClientPlayerEntity player = client.getInstance().player;
            GameMode gameMode = client.interactionManager == null ? null : client.interactionManager.getCurrentGameMode();
            if (player == null || gameMode == null || !gameMode.isSurvivalLike()) {
                noteBlocks = null;
                notePredictions.clear();
                tuned = false;
                tuneInitialUntunedBlocks = -1;
                manualTuningRequested = false;
                client.inGameHud.getChatHud().addMessage(Text.translatable(Main.MOD_ID+".player.invalid_state_tuning").formatted(Formatting.RED));
                return;
            }

            if (noteBlocks == null) {
                noteBlocks = new HashMap<>();

                HashMap<NoteBlockInstrument, ArrayList<BlockPos>> noteblocksForInstrument = new HashMap<>();
                for(NoteBlockInstrument instrument : NoteBlockInstrument.values())
                    noteblocksForInstrument.put(instrument, new ArrayList<>());
                final Vec3d playerEyePos = player.getEyePos();

                final int maxOffset;
                if(Main.config.expectedServerVersion == ModConfig.ExpectedServerVersion.v1_20_4_Or_Earlier) {
                    maxOffset = 7;
                }else if(Main.config.expectedServerVersion == ModConfig.ExpectedServerVersion.v1_20_5_Or_Later) {
                    maxOffset = (int) Math.ceil(player.getBlockInteractionRange() + 1.0 + 1.0);
                }else if(Main.config.expectedServerVersion == ModConfig.ExpectedServerVersion.All) {
                    maxOffset = Math.min(7, (int) Math.ceil(player.getBlockInteractionRange() + 1.0 + 1.0));
                }else {
                    throw new NotImplementedException("ExpectedServerVersion Value not implemented: " + Main.config.expectedServerVersion.name());
                }
                final ArrayList<Integer> orderedOffsets = new ArrayList<>();
                for(int offset = 0; offset <= maxOffset; offset++) {
                    orderedOffsets.add(offset);
                    if(offset != 0) orderedOffsets.add(offset * -1);
                }

                for(NoteBlockInstrument instrument : noteblocksForInstrument.keySet().toArray(new NoteBlockInstrument[0])) {
                    for (int y : orderedOffsets) {
                        for (int x : orderedOffsets) {
                            for (int z : orderedOffsets) {
                                Vec3d vec3d = playerEyePos.add(x, y, z);
                                BlockPos blockPos = new BlockPos(MathHelper.floor(vec3d.x), MathHelper.floor(vec3d.y), MathHelper.floor(vec3d.z));
                                if (!canInteractWith(player, blockPos))
                                    continue;
                                BlockState blockState = world.getBlockState(blockPos);
                                if (!blockState.isOf(Blocks.NOTE_BLOCK) || !world.isAir(blockPos.up()))
                                    continue;

                                if (blockState.get(Properties.INSTRUMENT) == instrument)
                                    noteblocksForInstrument.get(instrument).add(blockPos);
                            }
                        }
                    }
                }

                if(!instrumentMap.isEmpty()) {
                    HashMap<NoteBlockInstrument, ArrayList<BlockPos>> newNoteblocksForInstrument = new HashMap<>();
                    for(NoteBlockInstrument orig : noteblocksForInstrument.keySet()) {
                        NoteBlockInstrument mappedInstrument = instrumentMap.getOrDefault(orig, orig);
                        if(mappedInstrument == null) {
                            newNoteblocksForInstrument.put(orig, null);
                            continue;
                        }

                        newNoteblocksForInstrument.put(orig, noteblocksForInstrument.getOrDefault(instrumentMap.getOrDefault(orig, orig), new ArrayList<>()));
                    }
                    noteblocksForInstrument = newNoteblocksForInstrument;
                }

                ArrayList<Note> capturedNotes = new ArrayList<>();

                ArrayList<Note> neededNotes = new ArrayList<>();
                if (song != null) {
                    neededNotes.addAll(song.uniqueNotes);
                }
                if (Main.keyMappingManager != null) {
                    for (Note mappedNote : Main.keyMappingManager.getMappings().values()) {
                        if (mappedNote != null && !neededNotes.contains(mappedNote)) {
                            neededNotes.add(mappedNote);
                        }
                    }
                }


                for(Note note : neededNotes) {
                    ArrayList<BlockPos> availableBlocks = noteblocksForInstrument.get(note.instrument());
                    if(availableBlocks == null) {
                        getNotes(note.instrument()).put(note.note(), null);
                        continue;
                    }
                    BlockPos bestBlockPos = null;
                    int bestBlockTuningSteps = Integer.MAX_VALUE;
                    for(BlockPos blockPos : availableBlocks) {
                        boolean alreadyAssigned = false;
                        if (noteBlocks != null) {
                            for (HashMap<Byte, BlockPos> instrumentNotes : noteBlocks.values()) {
                                if (instrumentNotes != null && instrumentNotes.containsValue(blockPos)) {
                                    alreadyAssigned = true;
                                    break;
                                }
                            }
                        }
                        if (alreadyAssigned) continue;


                        int wantedNote = note.note();
                        BlockState blockState = world.getBlockState(blockPos);
                        if (!blockState.contains(Properties.NOTE)) continue;
                        int currentNote = blockState.get(Properties.NOTE);

                        int tuningSteps = wantedNote >= currentNote ? wantedNote - currentNote : (25 - currentNote) + wantedNote;

                        if(tuningSteps < bestBlockTuningSteps) {
                            bestBlockPos = blockPos;
                            bestBlockTuningSteps = tuningSteps;
                        }
                    }

                    if(bestBlockPos != null) {
                        capturedNotes.add(note);
                        availableBlocks.remove(bestBlockPos);
                        getNotes(note.instrument()).put(note.note(), bestBlockPos);
                    }
                }

                ArrayList<Note> missingNotes = new ArrayList<>(neededNotes);
                missingNotes.removeAll(capturedNotes);
                if (!missingNotes.isEmpty()) {
                    ChatHud chatHud = MinecraftClient.getInstance().inGameHud.getChatHud();
                    chatHud.addMessage(Text.translatable(Main.MOD_ID+".player.invalid_note_blocks").formatted(Formatting.RED));

                    HashMap<Block, Integer> missing = new HashMap<>();
                    for (Note note : missingNotes) {
                        NoteBlockInstrument mappedInstrument = instrumentMap.getOrDefault(note.instrument(), note.instrument());
                        if(mappedInstrument == null) continue;
                        Block block = Note.INSTRUMENT_BLOCKS.get(mappedInstrument);
                        Integer got = missing.get(block);
                        if (got == null) got = 0;
                        missing.put(block, got + 1);
                    }

                    missingInstrumentBlocks = missing;
                    missing.forEach((block, integer) -> chatHud.addMessage(Text.literal(block.getName().getString()+" × "+integer).formatted(Formatting.RED)));

                } else {
                    missingInstrumentBlocks.clear();
                }

            }


            int ping = 0;
            {
                PlayerListEntry playerListEntry;
                if (client.getNetworkHandler() != null && (playerListEntry = client.getNetworkHandler().getPlayerListEntry(client.player.getGameProfile().getId())) != null)
                    ping = playerListEntry.getLatency();
            }

            int fullyTunedBlocks = 0;
            HashMap<BlockPos, Integer> untunedNotes = new HashMap<>();

            if (noteBlocks != null) {
                for (HashMap<Byte, BlockPos> instrumentNotes : noteBlocks.values()) {
                    if (instrumentNotes == null) continue;
                    for (Map.Entry<Byte, BlockPos> entry : instrumentNotes.entrySet()) {
                        BlockPos blockPos = entry.getValue();
                        Byte wantedNote = entry.getKey();

                        if (blockPos == null) continue;

                        BlockState blockState = world.getBlockState(blockPos);
                        if (!blockState.contains(Properties.NOTE)) {
                            Main.LOGGER.warn("注意 {} 处音符盒在调整过程中改变了状态", blockPos);
                            noteBlocks = null;
                            tuned = false;
                            manualTuningRequested = false;
                            return;
                        }

                        int assumedNote = notePredictions.containsKey(blockPos) ? notePredictions.get(blockPos).getLeft() : blockState.get(Properties.NOTE); // blockState.get(Properties.NOTE) returns Integer

                        byte wantedNotePrimitive = wantedNote.byteValue();

                        if(assumedNote == wantedNotePrimitive && blockState.get(Properties.NOTE).intValue() == wantedNotePrimitive) {
                            fullyTunedBlocks++;
                        } else if (assumedNote != wantedNotePrimitive) {
                            untunedNotes.put(blockPos, blockState.get(Properties.NOTE).intValue());
                        }
                    }
                }
            }


            int existingUniqueNotesCount = 0;
            if (noteBlocks != null) {
                for(HashMap<Byte, BlockPos> instrumentNotes : noteBlocks.values()) {
                    if (instrumentNotes != null) {
                        for (BlockPos pos : instrumentNotes.values()) {
                            if (pos != null) {
                                existingUniqueNotesCount++;
                            }
                        }
                    }
                }
            }


            if(untunedNotes.isEmpty() && fullyTunedBlocks == existingUniqueNotesCount) {
                if(lastInteractAt == -1 || System.currentTimeMillis() - lastInteractAt >= ping * 2 + 100) {
                    tuned = true;
                    if (running && tick == 0 && index == 0) {
                        pausePlaybackUntil = System.currentTimeMillis() + (long) (Math.abs(Main.config.delayPlaybackStartBySecs) * 1000);
                    } else {
                        pausePlaybackUntil = -1L;
                    }
                    tuneInitialUntunedBlocks = -1;
                    manualTuningRequested = false;
                    client.inGameHud.getChatHud().addMessage(Text.translatable(Main.MOD_ID+".player.tuned").formatted(Formatting.GREEN));
                }
            } else {
                tuned = false;
                if(tuneInitialUntunedBlocks == -1 || tuneInitialUntunedBlocks < untunedNotes.size())
                    tuneInitialUntunedBlocks = untunedNotes.size();
                if (availableInteracts >= 1f && !untunedNotes.isEmpty()) {
                    BlockPos blockPosToTune = untunedNotes.keySet().iterator().next();
                    int currentNote = untunedNotes.get(blockPosToTune);
                    Byte targetNote = null;
                    if (noteBlocks != null) {
                        for (HashMap<Byte, BlockPos> instrumentNotes : noteBlocks.values()) {
                            if (instrumentNotes != null) {
                                for (Map.Entry<Byte, BlockPos> entry : instrumentNotes.entrySet()) {
                                    if (blockPosToTune.equals(entry.getValue())) {
                                        targetNote = entry.getKey();
                                        break;
                                    }
                                }
                            }
                            if (targetNote != null) break;
                        }
                    }

                    if (targetNote != null) {
                        byte targetNotePrimitive = targetNote.byteValue();
                        int tuningSteps = targetNotePrimitive >= currentNote ? targetNotePrimitive - currentNote : (25 - currentNote) + targetNotePrimitive;

                        if (tuningSteps > 0) {
                            int predictedNote = (currentNote + 1) % 25;
                            notePredictions.put(blockPosToTune, new Pair<>(predictedNote, System.currentTimeMillis() + ping * 2 + 100));
                            client.interactionManager.interactBlock(client.player, Hand.MAIN_HAND, new BlockHitResult(Vec3d.of(blockPosToTune), Direction.UP, blockPosToTune, false));
                            lastInteractAt = System.currentTimeMillis();
                            availableInteracts -= 1f;

                            client.player.swingHand(Hand.MAIN_HAND);

                            int totalUntuned = untunedNotes.size();
                            int tunedCount = existingUniqueNotesCount - totalUntuned;

                        } else {

                            untunedNotes.remove(blockPosToTune);
                        }
                    } else {

                        untunedNotes.remove(blockPosToTune); // Remove to avoid infinite loop
                    }
                } else if (!untunedNotes.isEmpty()) {

                }
            }
        }
        if((playbackThread == null || !playbackThread.isAlive()) && running && Main.config.disableAsyncPlayback) {
            try {
                tickPlayback();
            }catch (Exception ex) {
                ex.printStackTrace();
                stop();
            }
        }
    }

    private HashMap<Byte, BlockPos> getNotes(NoteBlockInstrument instrument) {
        if (noteBlocks == null) {
            noteBlocks = new HashMap<>();
        }
        return noteBlocks.computeIfAbsent(instrument, k -> new HashMap<>());
    }

    // Before 1.20.5, the server limits interacts to 6 Blocks from Player Eye to Block Center
    // With 1.20.5 and later, the server does a more complex check, to the closest point of a full block hitbox
    // (max distance is BlockInteractRange + 1.0).
    public boolean canInteractWith(ClientPlayerEntity player, BlockPos blockPos) {
        final Vec3d eyePos = player.getEyePos();
        if(Main.config.expectedServerVersion == ModConfig.ExpectedServerVersion.v1_20_4_Or_Earlier) {
            return eyePos.squaredDistanceTo(blockPos.toCenterPos()) <= 6.0 * 6.0;
        }else if(Main.config.expectedServerVersion == ModConfig.ExpectedServerVersion.v1_20_5_Or_Later) {
            double blockInteractRange = player.getBlockInteractionRange() + 1.0;
            return new Box(blockPos).squaredMagnitude(eyePos) < blockInteractRange * blockInteractRange;
        }else if(Main.config.expectedServerVersion == ModConfig.ExpectedServerVersion.All) {
            // Require both checks to succeed (aka use worst distance)
            double blockInteractRange = player.getBlockInteractionRange() + 1.0;
            return eyePos.squaredDistanceTo(blockPos.toCenterPos()) <= 6.0 * 6.0
                    && new Box(blockPos).squaredMagnitude(eyePos) < blockInteractRange * blockInteractRange;
        }else {
            throw new NotImplementedException("ExpectedServerVersion Value not implemented: " + Main.config.expectedServerVersion.name());
        }
    }

    public double getSongElapsedSeconds() {
        if(song == null) return 0;
        return song.ticksToMilliseconds(tick) / 1000;
    }
}
