package semmiedev.disc_jockey_revive.gui;

import net.minecraft.client.MinecraftClient;
import net.minecraft.client.gui.DrawContext;
import net.minecraft.client.gui.screen.narration.NarrationMessageBuilder;
import net.minecraft.client.gui.screen.narration.NarrationPart;
import net.minecraft.client.gui.widget.ButtonWidget;
import net.minecraft.client.gui.widget.EntryListWidget;
import net.minecraft.client.util.InputUtil;
import net.minecraft.text.Text;
import semmiedev.disc_jockey_revive.KeyMappingManager;
import semmiedev.disc_jockey_revive.Main;
import semmiedev.disc_jockey_revive.Note;
import semmiedev.disc_jockey_revive.gui.screen.EditKeyMappingsScreen;

import java.util.Comparator;
import java.util.Map;
import java.util.Objects;

public class KeyMappingListWidget extends EntryListWidget<KeyMappingListWidget.KeyMappingEntry> {

    private final EditKeyMappingsScreen parentScreen;
    private boolean buttonsActive = true; 

    public KeyMappingListWidget(MinecraftClient client, int width, int height, int top, int itemHeight, EditKeyMappingsScreen parentScreen) {
        super(client, width, height, top, itemHeight);
        this.parentScreen = parentScreen;
        this.centerListVertically = false; 
    }
    public void setMappings(Map<InputUtil.Key, Note> mappings) {
        this.clearEntries(); 
        for (Map.Entry<InputUtil.Key, Note> entry : mappings.entrySet()) {
            
            this.addEntry(new KeyMappingEntry(entry.getKey(), entry.getValue(), this.parentScreen));
        }
        
        this.children().sort(Comparator.comparing(entry -> entry.getKey().getTranslationKey()));
        
        setButtonsActive(this.buttonsActive);
    }
    public void setButtonsActive(boolean active) {
        this.buttonsActive = active;
        
        for (KeyMappingEntry entry : children()) {
            entry.setButtonsActive(active);
        }
    }
    @Override
    public int getRowWidth() {
        return this.width - 40; 
    }

    @Override
    protected int getScrollbarX() {
        return this.width - 10; 
    }

    @Override
    protected void appendClickableNarrations(NarrationMessageBuilder builder) {
    }

    public class KeyMappingEntry extends EntryListWidget.Entry<KeyMappingEntry> {
        private final InputUtil.Key key;
        private final Note note;
        private final EditKeyMappingsScreen screen;

        private ButtonWidget changeButton;
        private ButtonWidget removeButton;

        public KeyMappingEntry(InputUtil.Key key, Note note, EditKeyMappingsScreen screen) {
            this.key = key;
            this.note = note;
            this.screen = screen;
        }

        public InputUtil.Key getKey() {
            return key;
        }

        public Note getNote() {
            return note;
        }

        public void setButtonsActive(boolean active) {
            if (changeButton != null) changeButton.active = active;
            if (removeButton != null) removeButton.active = active;
        }
        @Override
        public void render(DrawContext context, int index, int y, int x, int entryWidth, int entryHeight, int mouseX, int mouseY, boolean hovered, float tickDelta) {
            MinecraftClient client = MinecraftClient.getInstance();
            int textY = y + (entryHeight - client.textRenderer.fontHeight) / 2;
            Text keyText = Text.translatable(key.getTranslationKey());
            context.drawTextWithShadow(client.textRenderer, keyText, x + 5, textY, 0xFFFFFF);
            String noteDisplayName = KeyMappingManager.getNoteDisplayName(note);
            context.drawTextWithShadow(client.textRenderer, noteDisplayName, x + 100, textY, 0xAAAAAA); 
            int buttonWidth = 50;
            int buttonHeight = 18;
            int buttonY = y + (entryHeight - buttonHeight) / 2;
            int buttonX = x + entryWidth - buttonWidth - 5; 
            changeButton = ButtonWidget.builder(Text.translatable(Main.MOD_ID + ".screen.edit_mappings.change"), button -> {
                screen.startWaitingForKeyPress(this); 
            }).dimensions(buttonX - buttonWidth - 5, buttonY, buttonWidth, buttonHeight).build(); 
            changeButton.render(context, mouseX, mouseY, tickDelta);
            changeButton.active = buttonsActive; 
            removeButton = ButtonWidget.builder(Text.translatable(Main.MOD_ID + ".screen.edit_mappings.remove"), button -> {
                screen.removeMapping(key); 
            }).dimensions(buttonX, buttonY, buttonWidth, buttonHeight).build();
            removeButton.render(context, mouseX, mouseY, tickDelta);
            removeButton.active = buttonsActive; 
        }

        @Override
        public boolean mouseClicked(double mouseX, double mouseY, int button) {
            
            if (changeButton.mouseClicked(mouseX, mouseY, button)) {
                return true;
            }
            if (removeButton.mouseClicked(mouseX, mouseY, button)) {
                return true;
            }
            return false; 
        }

        @Override
        public boolean isMouseOver(double mouseX, double mouseY) {
            return super.isMouseOver(mouseX, mouseY) ||
                    (changeButton != null && changeButton.isMouseOver(mouseX, mouseY)) ||
                    (removeButton != null && removeButton.isMouseOver(mouseX, mouseY));
        }
        public void appendClickableNarrations(NarrationMessageBuilder builder) {
            builder.put(NarrationPart.TITLE, Text.translatable(key.getTranslationKey()).append(" -> ").append(KeyMappingManager.getNoteDisplayName(note)));
            if (changeButton != null) changeButton.appendNarrations(builder);
            if (removeButton != null) removeButton.appendNarrations(builder);
        }
    }
}
