package semmiedev.disc_jockey_revive.gui.screen;

import net.minecraft.client.gui.DrawContext;
import net.minecraft.client.gui.screen.ConfirmScreen;
import net.minecraft.client.gui.screen.Screen;
import net.minecraft.client.gui.widget.ButtonWidget;
import net.minecraft.client.gui.widget.TextFieldWidget;
import net.minecraft.item.ItemStack;
import net.minecraft.text.MutableText;
import net.minecraft.text.Text;
import net.minecraft.util.Formatting;
import org.lwjgl.glfw.GLFW;
import semmiedev.disc_jockey_revive.Main;
import semmiedev.disc_jockey_revive.Note;
import semmiedev.disc_jockey_revive.Song;
import semmiedev.disc_jockey_revive.SongLoader;
import semmiedev.disc_jockey_revive.gui.SongListWidget;
import semmiedev.disc_jockey_revive.gui.hud.BlocksOverlay;
import semmiedev.disc_jockey_revive.gui.widget.ProgressBarWidget;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

import semmiedev.disc_jockey_revive.SongLoader.SongFolder;
import semmiedev.disc_jockey_revive.SongPlayer.PlayMode;
import org.jetbrains.annotations.Nullable;

public class DiscJockeyScreen extends Screen {
    private static final MutableText
            SELECT_SONG = Text.translatable(Main.MOD_ID+".screen.select_song"),
            PLAY = Text.translatable(Main.MOD_ID+".screen.play"),
            PLAY_STOP = Text.translatable(Main.MOD_ID+".screen.play.stop"),
            PREVIEW = Text.translatable(Main.MOD_ID+".screen.preview"),
            PREVIEW_STOP = Text.translatable(Main.MOD_ID+".screen.preview.stop"),
            DROP_HINT = Text.translatable(Main.MOD_ID+".screen.drop_hint").formatted(Formatting.GRAY)
                    ;

    private SongListWidget songListWidget;
    private ButtonWidget playButton, previewButton, prevButton, nextButton;
    public boolean shouldFilter;
    private String query = "";
    private ProgressBarWidget progressBar;
    private TextFieldWidget searchBar;

    private static final MutableText
            FOLDER_UP = Text.literal("↑"),
            CURRENT_FOLDER = Text.translatable(Main.MOD_ID+".screen.current_folder"),
            PLAY_MODE = Text.translatable(Main.MOD_ID+".screen.play_mode"),
            MODE_SINGLE = Text.translatable(Main.MOD_ID+".screen.mode_single"),
            MODE_LIST = Text.translatable(Main.MOD_ID+".screen.mode_list"),
            MODE_RANDOM = Text.translatable(Main.MOD_ID+".screen.mode_random"),
            MODE_STOP = Text.translatable(Main.MOD_ID+".screen.mode_stop");

    private static final MutableText
            OPEN_FOLDER = Text.translatable(Main.MOD_ID+".screen.open_folder"),
            RELOAD = Text.translatable(Main.MOD_ID+".screen.reload");

    private ButtonWidget folderUpButton, playModeButton;
    public SongFolder currentFolder;
    private PlayMode currentPlayMode = PlayMode.STOP_AFTER;

    public DiscJockeyScreen() {
        super(Main.NAME);
    }

    @Override
    protected void init() {
        shouldFilter = true;

        currentPlayMode = Main.config.playMode;

        if (!Main.config.currentFolderPath.isEmpty()) {
            currentFolder = findFolderByPath(Main.config.currentFolderPath);
            SongLoader.currentFolder = currentFolder;
        }


        int listTop = 40;
        int listBottom = height - 100;
        int listHeight = listBottom - listTop;
        songListWidget = new SongListWidget(client, width, height - 100, 40, 20, this);
        addDrawableChild(songListWidget);


        int playbackButtonHeight = 20;
        int playbackButtonY = listBottom +50;
        int centerX = width / 2;

        // 上一首
        prevButton = ButtonWidget.builder(Text.literal("⏮"), button -> playPreviousSong())
                .dimensions(centerX - 80, playbackButtonY, 40, playbackButtonHeight).build();
        addDrawableChild(prevButton);

        // 播放
        playButton = ButtonWidget.builder(PLAY, button -> {
            if (Main.SONG_PLAYER.running) {
                Main.SONG_PLAYER.stop();
            } else {
                SongListWidget.SongEntry entry = songListWidget.getSelectedSongOrNull();
                if (entry != null) {
                    Main.SONG_PLAYER.start(entry.song);
                }
            }
        }).dimensions(centerX - 30, playbackButtonY, 60, playbackButtonHeight).build();
        addDrawableChild(playButton);

        // 下一首
        nextButton = ButtonWidget.builder(Text.literal("⏭"), button -> playNextSong())
                .dimensions(centerX + 40, playbackButtonY, 40, playbackButtonHeight).build();
        addDrawableChild(nextButton);

        // 预览
        previewButton = ButtonWidget.builder(PREVIEW, button -> {
            if (Main.PREVIEWER.running) {
                Main.PREVIEWER.stop();
            } else {
                SongListWidget.SongEntry entry = songListWidget.getSelectedSongOrNull();
                if (entry != null) Main.PREVIEWER.start(entry.song);
            }
        }).dimensions(centerX + 90, playbackButtonY, 60, playbackButtonHeight).build();
        addDrawableChild(previewButton);

        // 进度条
        int progressBarHeight = 10;
        int progressBarY = playbackButtonY + playbackButtonHeight + 5;
        progressBar = new ProgressBarWidget(centerX - 100, progressBarY, 200, progressBarHeight,
                Text.translatable(Main.MOD_ID + ".progress"), 0, 100) {
            @Override
            protected void onProgressChanged(double progress) {
                if (Main.SONG_PLAYER.running && Main.SONG_PLAYER.song != null) {
                    double targetTick = progress * Main.SONG_PLAYER.song.length / 100.0;
                    Main.SONG_PLAYER.tick = targetTick;
                    updateNoteIndexFromTick();
                }
            }
        };
        addDrawableChild(progressBar);

        // 搜索框
        int searchBarWidth = 150;
        int searchBarHeight = 20;
        int searchBarX = 10;
        int searchBarY = height - searchBarHeight - 10;
        searchBar = new TextFieldWidget(textRenderer, searchBarX, searchBarY, searchBarWidth, searchBarHeight,
                Text.translatable(Main.MOD_ID+".screen.search"));
        searchBar.setChangedListener(query -> {
            query = query.toLowerCase().replaceAll("\\s", "");
            if (this.query.equals(query)) return;
            this.query = query;
            shouldFilter = true;
        });
        addDrawableChild(searchBar);
        int otherButtonWidth = 100;
        int otherButtonHeight = 20;
        int otherButtonMargin = 10;

        // Blocks
        int blocksButtonX = width - otherButtonWidth - otherButtonMargin;
        int blocksButtonY = height - otherButtonHeight - otherButtonMargin - otherButtonHeight - otherButtonMargin;
        addDrawableChild(ButtonWidget.builder(Text.translatable(Main.MOD_ID+".screen.blocks"), button -> {
            if (BlocksOverlay.itemStacks == null) {
                SongListWidget.SongEntry entry = songListWidget.getSelectedSongOrNull();
                if (entry != null) {
                    client.setScreen(null);
                    updateBlocksOverlay(entry.song);
                }
            } else {
                BlocksOverlay.itemStacks = null;
                client.setScreen(null);
            }
        }).dimensions(blocksButtonX, blocksButtonY, otherButtonWidth, otherButtonHeight).build());

        // 打开文件夹
        int openFolderButtonX = width - otherButtonWidth - otherButtonMargin;
        int openFolderButtonY = height - otherButtonHeight - otherButtonMargin - otherButtonHeight - otherButtonMargin + otherButtonHeight + otherButtonMargin; // 在 Blocks 按钮下方一行
        addDrawableChild(ButtonWidget.builder(OPEN_FOLDER, button -> openSongsFolder())
                .dimensions(openFolderButtonX, openFolderButtonY, otherButtonWidth, otherButtonHeight).build());

        // 重新加载
        int reloadButtonX = width - otherButtonWidth - otherButtonMargin - otherButtonWidth - otherButtonMargin;
        int reloadButtonY = openFolderButtonY;
        addDrawableChild(ButtonWidget.builder(RELOAD, button -> {
            SongLoader.loadSongs();
            client.setScreen(null);
        }).dimensions(reloadButtonX, reloadButtonY, otherButtonWidth, otherButtonHeight).build());



        folderUpButton = ButtonWidget.builder(FOLDER_UP, button -> {
            if (currentFolder != null) {
                SongFolder parent = findParentFolder(currentFolder);
                currentFolder = parent;
                SongLoader.currentFolder = parent;
                shouldFilter = true;
            }
        }).dimensions(10, 10, 20, 20).build();
        addDrawableChild(folderUpButton);


        playModeButton = ButtonWidget.builder(getPlayModeText(), button -> {
            switch (currentPlayMode) {
                case SINGLE_LOOP -> currentPlayMode = PlayMode.LIST_LOOP;
                case LIST_LOOP -> currentPlayMode = PlayMode.RANDOM;
                case RANDOM -> currentPlayMode = PlayMode.STOP_AFTER;
                case STOP_AFTER -> currentPlayMode = PlayMode.SINGLE_LOOP;
            }
            Main.SONG_PLAYER.setPlayMode(currentPlayMode);
            playModeButton.setMessage(getPlayModeText());
        }).dimensions(width - 120, 10, 100, 20).build();
        addDrawableChild(playModeButton);

        for (int i = 0; i < SongLoader.SONGS.size(); i++) {
            Song song = SongLoader.SONGS.get(i);
            song.entry.songListWidget = songListWidget;
            if (song.entry.selected) songListWidget.setSelected(song.entry);

            if (song.folder != null) {
                boolean folderEntryExists = SongLoader.FOLDERS.stream()
                        .anyMatch(f -> f == song.folder || findFolderByPathInSubfolders(f, song.folder.path) != null);
                if (!folderEntryExists) {
                } else {
                    if (song.folder.entry == null) {
                        song.folder.entry = new SongListWidget.FolderEntry(song.folder, songListWidget);
                    }
                }
            }
        }
    }

    // 播放进度
    private void updateNoteIndexFromTick() {
        if (Main.SONG_PLAYER.song == null) return;

        double targetTick = Main.SONG_PLAYER.tick;
        int newIndex = 0;
        for (int i = 0; i < Main.SONG_PLAYER.song.notes.length; i++) {
            long note = Main.SONG_PLAYER.song.notes[i];
            if ((short)note > targetTick) {
                newIndex = i;
                break;
            }
            newIndex = i;
        }
        Main.SONG_PLAYER.index = newIndex;
    }

    private void updateBlocksOverlay(Song song) {
        BlocksOverlay.itemStacks = new ItemStack[0];
        BlocksOverlay.amounts = new int[0];
        BlocksOverlay.amountOfNoteBlocks = song.uniqueNotes.size();

        for (Note note : song.uniqueNotes) {
            ItemStack itemStack = Note.INSTRUMENT_BLOCKS.get(note.instrument()).asItem().getDefaultStack();
            int index = -1;

            for (int i = 0; i < BlocksOverlay.itemStacks.length; i++) {
                if (BlocksOverlay.itemStacks[i].getItem() == itemStack.getItem()) {
                    index = i;
                    break;
                }
            }

            if (index == -1) {
                BlocksOverlay.itemStacks = Arrays.copyOf(BlocksOverlay.itemStacks, BlocksOverlay.itemStacks.length + 1);
                BlocksOverlay.amounts = Arrays.copyOf(BlocksOverlay.amounts, BlocksOverlay.amounts.length + 1);

                BlocksOverlay.itemStacks[BlocksOverlay.itemStacks.length - 1] = itemStack;
                BlocksOverlay.amounts[BlocksOverlay.amounts.length - 1] = 1;
            } else {
                BlocksOverlay.amounts[index] = BlocksOverlay.amounts[index] + 1;
            }
        }
    }

    // 打开文件夹
    private void openSongsFolder() {
        try {
            String folderPath = currentFolder != null ?
                    currentFolder.path : Main.songsFolder.getAbsolutePath();

            File target = new File(folderPath);
            if (!target.exists()) {
                client.inGameHud.getChatHud().addMessage(
                        Text.translatable(Main.MOD_ID+".screen.folder_not_exist", folderPath)
                                .formatted(Formatting.RED));
                return;
            }

            if (System.getProperty("os.name").toLowerCase().contains("win")) {
                new ProcessBuilder("explorer.exe", "/select,", target.getAbsolutePath()).start();
            } else {
                java.awt.Desktop.getDesktop().open(target);
            }
        } catch (Exception e) {
            Main.LOGGER.error("打开文件夹失败", e);
            client.inGameHud.getChatHud().addMessage(
                    Text.translatable(Main.MOD_ID+".screen.open_folder_failed")
                            .formatted(Formatting.RED));
        }
    }

    // 预览
    private void playPreviousSong() {
        if (SongLoader.currentFolder == null || SongLoader.currentFolder.songs.isEmpty()) return;

        int currentIndex = SongLoader.currentFolder.songs.indexOf(Main.SONG_PLAYER.song);
        if (currentIndex == -1) {
            if (!SongLoader.currentFolder.songs.isEmpty()) {
                startSong(SongLoader.currentFolder.songs.get(0));
            }
            return;
        }

        int prevIndex = (currentIndex - 1 + SongLoader.currentFolder.songs.size()) % SongLoader.currentFolder.songs.size();
        startSong(SongLoader.currentFolder.songs.get(prevIndex));
    }

    // 下一首
    private void playNextSong() {
        if (SongLoader.currentFolder == null || SongLoader.currentFolder.songs.isEmpty()) return;

        if (Main.SONG_PLAYER.playMode == PlayMode.RANDOM) {
            int newIndex;
            int currentIndex = SongLoader.currentFolder.songs.indexOf(Main.SONG_PLAYER.song);
            do {
                newIndex = (int)(Math.random() * SongLoader.currentFolder.songs.size());
            } while (newIndex == currentIndex &&
                    SongLoader.currentFolder.songs.size() > 1);
            startSong(SongLoader.currentFolder.songs.get(newIndex));
        } else {
            int currentIndex = SongLoader.currentFolder.songs.indexOf(Main.SONG_PLAYER.song);
            if (currentIndex == -1) {
                if (!SongLoader.currentFolder.songs.isEmpty()) {
                    startSong(SongLoader.currentFolder.songs.get(0));
                }
                return;
            }

            int nextIndex = (currentIndex + 1) % SongLoader.currentFolder.songs.size();
            startSong(SongLoader.currentFolder.songs.get(nextIndex));
        }
    }

    private void startSong(Song song) {
        SongListWidget.SongEntry entry = findEntryForSong(song);
        if (entry != null) {
            songListWidget.setSelected(entry);
            Main.SONG_PLAYER.start(song);
        }
    }

    private SongListWidget.SongEntry findEntryForSong(Song song) {
        for (SongListWidget.Entry entry : songListWidget.children()) {
            if (entry instanceof SongListWidget.SongEntry songEntry && songEntry.song == song) {
                return songEntry;
            }
        }
        return null;
    }


    @Override
    public void render(DrawContext context, int mouseX, int mouseY, float delta) {
        renderBackground(context, mouseX, mouseY, delta);
        super.render(context, mouseX, mouseY, delta);

        // 标题
        context.drawCenteredTextWithShadow(textRenderer, DROP_HINT, width / 2, 5, 0xFFFFFF);
        context.drawCenteredTextWithShadow(textRenderer, SELECT_SONG, width / 2, 20, 0xFFFFFF);

        // 文件夹 播放模式
        String folderName = currentFolder == null ? "/" : currentFolder.name;
        context.drawTextWithShadow(textRenderer, CURRENT_FOLDER.getString() + ": " + folderName, 35, 15, 0xFFFFFF);
        context.drawTextWithShadow(textRenderer, PLAY_MODE.getString() + ":", width - 220, 15, 0xFFFFFF);

        // 进度条
        if (Main.SONG_PLAYER.running && Main.SONG_PLAYER.song != null) {
            double progress = (Main.SONG_PLAYER.tick / Main.SONG_PLAYER.song.length) * 100;
            if (!progressBar.isDragging()) {
                progressBar.setProgress(progress);
            }
            String timeText = formatTime(Main.SONG_PLAYER.getSongElapsedSeconds()) + " / " +
                    formatTime(Main.SONG_PLAYER.song.getLengthInSeconds());
            int timeTextY = progressBar.getY() + progressBar.getHeight() + 2;
            context.drawTextWithShadow(textRenderer, timeText, width / 2 - textRenderer.getWidth(timeText) / 2, timeTextY, 0xFFFFFF);
        } else {
            progressBar.setProgress(0);
            String timeText = formatTime(0) + " / " + formatTime(0);
            int timeTextY = progressBar.getY() + progressBar.getHeight() + 2;
            context.drawTextWithShadow(textRenderer, timeText, width / 2 - textRenderer.getWidth(timeText) / 2, timeTextY, 0xFFFFFF);
        }
    }

    // 时间格式hua
    private String formatTime(double seconds) {
        int totalSeconds = (int) seconds;
        int minutes = totalSeconds / 60;
        int secs = totalSeconds % 60;
        return String.format("%02d:%02d", minutes, secs);
    }

    @Override
    public void tick() {
        super.tick();

        previewButton.setMessage(Main.PREVIEWER.running ? PREVIEW_STOP : PREVIEW);
        playButton.setMessage(Main.SONG_PLAYER.running ? PLAY_STOP : PLAY);

        if (shouldFilter) {
            shouldFilter = false;
            songListWidget.children().clear();
            boolean empty = query.isEmpty();

            boolean isInSongsOrSubfolder = currentFolder == null ||
                    (Main.songsFolder != null && currentFolder.path.startsWith(Main.songsFolder.getPath()));

            if (currentFolder == null) {
                for (SongFolder folder : SongLoader.FOLDERS) {
                    if (empty || folder.name.toLowerCase().contains(query)) {
                        if (folder.entry == null) {
                            folder.entry = new SongListWidget.FolderEntry(folder, songListWidget);
                        }
                        songListWidget.children().add(folder.entry);
                    }
                }
            } else {
                SongListWidget.FolderEntry parentEntry = new SongListWidget.FolderEntry(null, songListWidget);
                parentEntry.displayName = "..";
                songListWidget.children().add(parentEntry);

                for (SongFolder subFolder : currentFolder.subFolders) {
                    if (empty || subFolder.name.toLowerCase().contains(query)) {
                        if (subFolder.entry == null) {
                            subFolder.entry = new SongListWidget.FolderEntry(subFolder, songListWidget);
                        }
                        songListWidget.children().add(subFolder.entry);
                    }
                }
            }

            if (isInSongsOrSubfolder) {
                List<Song> songsToShow = currentFolder == null ?
                        SongLoader.SONGS.stream()
                                .filter(song -> song.folder == null)
                                .collect(Collectors.toList()) :
                        currentFolder.songs.stream()
                                .filter(song -> song.folder == currentFolder)
                                .collect(Collectors.toList());

                // 已收藏
                for (Song song : songsToShow) {
                    if (song.entry.favorite && (empty || song.searchableFileName.contains(query) || song.searchableName.contains(query))) {
                        songListWidget.children().add(song.entry);
                    }
                }

                // 未收藏
                for (Song song : songsToShow) {
                    if (!song.entry.favorite && (empty || song.searchableFileName.contains(query) || song.searchableName.contains(query))) {
                        songListWidget.children().add(song.entry);
                    }
                }
            }
        }
    }

    // 拖文件
    @Override
    public void onFilesDropped(List<Path> paths) {
        String string = paths.stream().map(Path::getFileName).map(Path::toString).collect(Collectors.joining(", "));
        if (string.length() > 300) string = string.substring(0, 300)+"...";

        client.setScreen(new ConfirmScreen(confirmed -> {
            if (confirmed) {
                paths.forEach(path -> {
                    try {
                        File file = path.toFile();

                        if (SongLoader.SONGS.stream().anyMatch(input -> input.fileName.equalsIgnoreCase(file.getName()))) return;

                        Song song = SongLoader.loadSong(file);
                        if (song != null) {
                            File targetFolder = currentFolder != null ? new File(currentFolder.path) : Main.songsFolder;
                            Files.copy(path, targetFolder.toPath().resolve(file.getName()));
                            SongLoader.loadSongs();
                        }
                    } catch (IOException exception) {
                        Main.LOGGER.warn("无法将歌曲文件从 {} 复制到 {} ", path, Main.songsFolder.toPath(), exception);
                    }
                });
            }
            client.setScreen(this);
        }, Text.translatable(Main.MOD_ID+".screen.drop_confirm"), Text.literal(string)));
    }

    @Override
    public boolean shouldPause() {
        return false;
    }

    @Override
    public void close() {
        super.close();
        Main.config.currentFolderPath = currentFolder != null ? currentFolder.path : "";
        Main.config.playMode = currentPlayMode;
        new Thread(() -> Main.configHolder.save()).start();
    }

    private SongFolder findFolderByPath(String path) {
        for (SongFolder folder : SongLoader.FOLDERS) {
            if (folder.path.equals(path)) {
                return folder;
            }
        }

        for (SongFolder folder : SongLoader.FOLDERS) {
            SongFolder found = findFolderByPathInSubfolders(folder, path);
            if (found != null) {
                return found;
            }
        }
        return null;
    }

    @Nullable
    private SongFolder findFolderByPathInSubfolders(SongFolder parent, String targetPath) {
        for (SongFolder subFolder : parent.subFolders) {
            if (subFolder.path.equals(targetPath)) {
                return subFolder;
            }
            SongFolder found = findFolderByPathInSubfolders(subFolder, targetPath);
            if (found != null) {
                return found;
            }
        }
        return null;
    }

    @Nullable
    public SongFolder findParentFolder(@Nullable SongFolder targetFolder) {
        if (targetFolder == null) {
            return null;
        }
        for (SongFolder folder : SongLoader.FOLDERS) {
            if (folder.subFolders.contains(targetFolder)) {
                return folder;
            }
            SongFolder parentInSub = findParentFolderInSubfoldersForParent(folder, targetFolder);
            if (parentInSub != null) {
                return parentInSub;
            }
        }
        return null;
    }

    @Nullable
    private SongFolder findParentFolderInSubfoldersForParent(SongFolder current, SongFolder targetFolder) {
        for (SongFolder subFolder : current.subFolders) {
            if (subFolder.subFolders.contains(targetFolder)) {
                return subFolder;
            }
            SongFolder parentInSub = findParentFolderInSubfoldersForParent(subFolder, targetFolder);
            if (parentInSub != null) {
                return parentInSub;
            }
        }
        return null;
    }


    private Text getPlayModeText() {
        return switch (currentPlayMode) {
            case SINGLE_LOOP -> MODE_SINGLE;
            case LIST_LOOP -> MODE_LIST;
            case RANDOM -> MODE_RANDOM;
            case STOP_AFTER -> MODE_STOP;
        };
    }
}
