package semmiedev.disc_jockey_revive.gui.screen;

import net.minecraft.client.MinecraftClient;
import net.minecraft.client.gui.DrawContext;
import net.minecraft.client.gui.Drawable;
import net.minecraft.client.gui.screen.Screen;
import net.minecraft.client.gui.widget.ButtonWidget;
import net.minecraft.client.util.InputUtil;
import net.minecraft.text.Text;
import org.lwjgl.glfw.GLFW;
import semmiedev.disc_jockey_revive.Main;
import semmiedev.disc_jockey_revive.Note;
import semmiedev.disc_jockey_revive.gui.KeyMappingListWidget;

import java.util.Map;

public class EditKeyMappingsScreen extends Screen {

    private static final Text TITLE = Text.translatable(Main.MOD_ID + ".screen.edit_mappings.title");
    private static final Text ADD_MAPPING_BUTTON_TEXT = Text.translatable(Main.MOD_ID + ".screen.edit_mappings.add_mapping");
    private static final Text DONE_BUTTON_TEXT = Text.translatable("gui.done");
    private static final Text PRESS_KEY_INSTRUCTION = Text.translatable(Main.MOD_ID + ".screen.edit_mappings.press_key");

    private final Screen parent;
    private KeyMappingListWidget mappingListWidget;
    private ButtonWidget addMappingButton;
    private ButtonWidget doneButton;

    private boolean waitingForKeyPress = false;
    private KeyMappingListWidget.KeyMappingEntry entryToEdit = null; 

    public EditKeyMappingsScreen(Screen parent) {
        super(TITLE);
        this.parent = parent;
    }

    @Override
    protected void init() {
        super.init();

        int listTop = 40;
        int listBottom = this.height - 50; 
        int listHeight = listBottom - listTop;

        mappingListWidget = new KeyMappingListWidget(this.client, this.width, listHeight, listTop, 20, this);
        addDrawableChild(mappingListWidget);
        refreshMappingList();
        int buttonWidth = 100;
        int buttonHeight = 20;
        int buttonY = this.height - 30;
        int buttonX = this.width / 2 - buttonWidth - 5; 

        addMappingButton = ButtonWidget.builder(ADD_MAPPING_BUTTON_TEXT, button -> {
            startWaitingForKeyPress(null); 
        }).dimensions(buttonX, buttonY, buttonWidth, buttonHeight).build();
        addDrawableChild(addMappingButton);
        buttonX = this.width / 2 + 5; 
        doneButton = ButtonWidget.builder(DONE_BUTTON_TEXT, button -> {
            Main.keyMappingManager.saveMappings(); 
            this.client.setScreen(this.parent); 
        }).dimensions(buttonX, buttonY, buttonWidth, buttonHeight).build();
        addDrawableChild(doneButton);
    }

    private void refreshMappingList() {
        
        mappingListWidget.setMappings(Main.keyMappingManager.getMappings());
    }
    public void startWaitingForKeyPress(KeyMappingListWidget.KeyMappingEntry entry) {
        this.waitingForKeyPress = true;
        this.entryToEdit = entry; 
        
        addMappingButton.active = false;
        doneButton.active = false;
        mappingListWidget.setButtonsActive(false);
    }
    public void addNewMapping(InputUtil.Key key, Note note) {
        Main.keyMappingManager.setMapping(key, note);
        refreshMappingList();
        
    }
    public void removeMapping(InputUtil.Key key) {
        Main.keyMappingManager.removeMapping(key);
        refreshMappingList();
    }
    public void stopWaitingForKeyPress() {
        this.waitingForKeyPress = false;
        this.entryToEdit = null;
        
        addMappingButton.active = true;
        doneButton.active = true;
        mappingListWidget.setButtonsActive(true);
    }

    @Override
    public boolean keyPressed(int keyCode, int scanCode, int modifiers) {
        if (waitingForKeyPress) {
            if (keyCode == GLFW.GLFW_KEY_ESCAPE) {
                
                stopWaitingForKeyPress();
                return true;
            }

            InputUtil.Key pressedKey = InputUtil.fromKeyCode(keyCode, scanCode);

            if (entryToEdit != null) {
                
                Note note = entryToEdit.getNote();
                
                Main.keyMappingManager.removeMapping(entryToEdit.getKey());
                
                Main.keyMappingManager.setMapping(pressedKey, note);
                refreshMappingList();
                stopWaitingForKeyPress();
            } else {
                this.client.setScreen(new SelectNoteScreen(this, pressedKey));
                
            }
            return true; 
        }
        if (mappingListWidget.keyPressed(keyCode, scanCode, modifiers)) {
            return true;
        }
        if (keyCode == GLFW.GLFW_KEY_ESCAPE) {
            Main.keyMappingManager.saveMappings();
            this.client.setScreen(this.parent);
            return true;
        }

        return super.keyPressed(keyCode, scanCode, modifiers);
    }

    @Override
    public void render(DrawContext context, int mouseX, int mouseY, float delta) {
        super.render(context, mouseX, mouseY, delta);
        context.drawCenteredTextWithShadow(textRenderer, TITLE, this.width / 2, 10, 0xFFFFFF);
        if (waitingForKeyPress) {
            
            context.fill(0, 0, this.width, this.height, 0x80000000); 
            context.drawCenteredTextWithShadow(textRenderer, PRESS_KEY_INSTRUCTION, this.width / 2, this.height / 2 - 10, 0xFFFFFF);
        }
        mappingListWidget.render(context, mouseX, mouseY, delta);
    }

    @Override
    public boolean shouldPause() {
        
        return true;
    }
}
