package semmiedev.disc_jockey_revive.gui.screen;

import me.shedaniel.cloth.clothconfig.shadowed.blue.endless.jankson.annotation.Nullable;
import net.minecraft.client.MinecraftClient;
import net.minecraft.client.gui.DrawContext;
import net.minecraft.client.gui.screen.Screen;
import net.minecraft.client.gui.widget.ButtonWidget;
import net.minecraft.client.util.InputUtil;
import net.minecraft.sound.SoundCategory;
import net.minecraft.text.Text;
import net.minecraft.util.math.MathHelper;
import net.minecraft.util.math.Vec3d;
import semmiedev.disc_jockey_revive.Main;
import semmiedev.disc_jockey_revive.Note;
import semmiedev.disc_jockey_revive.KeyMappingManager;
import org.lwjgl.glfw.GLFW;

import java.util.HashMap;
import java.util.Map;

public class LiveDjScreen extends Screen {

    private static final Text TITLE = Text.translatable(Main.MOD_ID + ".screen.live_dj.title");
    private static final Text INSTRUCTIONS = Text.translatable(Main.MOD_ID + ".screen.live_dj.instructions");
    private static final Text EDIT_MAPPINGS_BUTTON_TEXT = Text.translatable(Main.MOD_ID + ".screen.live_dj.edit_mappings");
    private static final Text START_TUNING_BUTTON_TEXT = Text.translatable(Main.MOD_ID + ".screen.live_dj.start_tuning"); 
    private static final Text NOT_TUNED_MESSAGE = Text.translatable(Main.MOD_ID + ".player.not_tuned").formatted(net.minecraft.util.Formatting.RED);
    private static final Text NOTE_BLOCK_MISSING_MESSAGE = Text.translatable(Main.MOD_ID + ".player.note_block_missing_live").formatted(net.minecraft.util.Formatting.RED);
    private static final Text TOO_FAR_MESSAGE = Text.translatable(Main.MOD_ID + ".player.to_far_live").formatted(net.minecraft.util.Formatting.RED);
    private static final Text RATE_LIMITED_MESSAGE = Text.translatable(Main.MOD_ID + ".player.rate_limited_live").formatted(net.minecraft.util.Formatting.YELLOW);
    private ButtonWidget startTuningButton; 

    public LiveDjScreen() {
        super(TITLE);
    }

    @Override
    protected void init() {
        super.init();

        int centerX = this.width / 2;
        int buttonWidth = 150;
        int buttonHeight = 20;
        int buttonY = this.height - 30; 
        int margin = 5;
        addDrawableChild(ButtonWidget.builder(EDIT_MAPPINGS_BUTTON_TEXT, button -> {
            MinecraftClient.getInstance().setScreen(new EditKeyMappingsScreen(this));
        }).dimensions(centerX - buttonWidth - margin, buttonY, buttonWidth, buttonHeight).build());
        startTuningButton = ButtonWidget.builder(START_TUNING_BUTTON_TEXT, button -> {
            Main.SONG_PLAYER.startTuning(); 
        }).dimensions(centerX + margin, buttonY, buttonWidth, buttonHeight).build();
        addDrawableChild(startTuningButton);
    }

    @Override
    public void render(DrawContext context, int mouseX, int mouseY, float delta) {
        super.render(context, mouseX, mouseY, delta); 
        context.drawCenteredTextWithShadow(textRenderer, TITLE, this.width / 2, 10, 0xFFFFFF);
        context.drawCenteredTextWithShadow(textRenderer, INSTRUCTIONS, this.width / 2, 30, 0xFFFFFF);
        Text tuningStatusText;
        if (Main.SONG_PLAYER.noteBlocks == null) {
            tuningStatusText = Text.translatable(Main.MOD_ID + ".player.discovering").formatted(net.minecraft.util.Formatting.YELLOW);
            startTuningButton.active = true; 
            startTuningButton.visible = true;
        } else if (!Main.SONG_PLAYER.tuned) {
            
            int totalNeeded = 0;
            if (Main.SONG_PLAYER.noteBlocks != null) {
                for (HashMap<Byte, net.minecraft.util.math.BlockPos> instrumentNotes : Main.SONG_PLAYER.noteBlocks.values()) {
                    if (instrumentNotes != null) {
                        for (net.minecraft.util.math.BlockPos pos : instrumentNotes.values()) {
                            if (pos != null) {
                                totalNeeded++;
                            }
                        }
                    }
                }
            }
            int tunedCount = 0;
            
            MinecraftClient client = MinecraftClient.getInstance();
            if (client.world != null && Main.SONG_PLAYER.noteBlocks != null) {
                for (HashMap<Byte, net.minecraft.util.math.BlockPos> instrumentNotes : Main.SONG_PLAYER.noteBlocks.values()) {
                    if (instrumentNotes != null) {
                        for (Map.Entry<Byte, net.minecraft.util.math.BlockPos> entry : instrumentNotes.entrySet()) {
                            net.minecraft.util.math.BlockPos pos = entry.getValue();
                            Byte wantedNote = entry.getKey();
                            if (pos != null) {
                                
                                if (client.world.getBlockState(pos).contains(net.minecraft.state.property.Properties.NOTE)) {
                                    int currentNote = client.world.getBlockState(pos).get(net.minecraft.state.property.Properties.NOTE);
                                    if (currentNote == wantedNote.byteValue()) { 
                                        tunedCount++;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            if (totalNeeded > 0) {
                tuningStatusText = Text.translatable(Main.MOD_ID + ".player.tuning_progress", tunedCount, totalNeeded).formatted(net.minecraft.util.Formatting.YELLOW);
            } else {
                
                tuningStatusText = Text.translatable(Main.MOD_ID + ".player.finding_blocks").formatted(net.minecraft.util.Formatting.YELLOW);
            }
            startTuningButton.active = true; 
            startTuningButton.visible = true;

        } else {
            tuningStatusText = Text.translatable(Main.MOD_ID + ".player.tuned").formatted(net.minecraft.util.Formatting.GREEN);
            startTuningButton.active = false; 
            startTuningButton.visible = false; 
        }
        context.drawCenteredTextWithShadow(textRenderer, tuningStatusText, this.width / 2, 50, 0xFFFFFF);
    }

    @Override
    public boolean keyPressed(int keyCode, int scanCode, int modifiers) {
        
        InputUtil.Key key = InputUtil.fromKeyCode(keyCode, scanCode);
        if (keyCode == GLFW.GLFW_KEY_ESCAPE) {
            this.close();
            return true;
        }
        Note note = Main.keyMappingManager.getNoteForKey(key);

        if (note != null) {
            
            if (!Main.SONG_PLAYER.tuned) {
                MinecraftClient.getInstance().inGameHud.getChatHud().addMessage(NOT_TUNED_MESSAGE);
                return true; 
            }
            boolean played = Main.SONG_PLAYER.playNoteBlock(note);
            if (!played) {
                @Nullable net.minecraft.util.math.BlockPos blockPos = null;
                if (Main.SONG_PLAYER.noteBlocks != null && Main.SONG_PLAYER.noteBlocks.containsKey(note.instrument())) {
                    blockPos = Main.SONG_PLAYER.noteBlocks.get(note.instrument()).get(note.note());
                }

                if (blockPos == null) {
                    MinecraftClient.getInstance().inGameHud.getChatHud().addMessage(NOTE_BLOCK_MISSING_MESSAGE.copy().append(" (" + KeyMappingManager.getNoteDisplayName(note) + ")"));
                } else if (!Main.SONG_PLAYER.canInteractWith(MinecraftClient.getInstance().player, blockPos)) {
                    MinecraftClient.getInstance().inGameHud.getChatHud().addMessage(TOO_FAR_MESSAGE.copy().append(" (" + KeyMappingManager.getNoteDisplayName(note) + ")"));
                } else {
                    
                    MinecraftClient.getInstance().inGameHud.getChatHud().addMessage(RATE_LIMITED_MESSAGE);
                }
            }

            return true; 
        }
        return super.keyPressed(keyCode, scanCode, modifiers);
    }

    @Override
    public void tick() {
        super.tick();
        
        if (startTuningButton != null) {
            startTuningButton.visible = !Main.SONG_PLAYER.tuned;
            startTuningButton.active = !Main.SONG_PLAYER.isTuningEnabled(); 
        }
    }
    @Override
    public boolean shouldPause() {
        return false; 
    }

    @Override
    public void close() {
        
        super.close();
    }
}
