package semmiedev.disc_jockey_revive.gui.screen;

import net.minecraft.block.enums.NoteBlockInstrument;
import net.minecraft.client.MinecraftClient;
import net.minecraft.client.gui.DrawContext;
import net.minecraft.client.gui.screen.Screen;
import net.minecraft.client.gui.widget.ButtonWidget;
import net.minecraft.client.gui.widget.CyclingButtonWidget;
import net.minecraft.client.gui.widget.SliderWidget;
import net.minecraft.client.util.InputUtil;
import net.minecraft.sound.SoundCategory;
import net.minecraft.sound.SoundEvent;
import net.minecraft.text.Text;
import net.minecraft.util.Identifier;
import net.minecraft.util.math.MathHelper;
import net.minecraft.util.math.Vec3d;
import org.lwjgl.glfw.GLFW;
import semmiedev.disc_jockey_revive.Main;
import semmiedev.disc_jockey_revive.Note;
import semmiedev.disc_jockey_revive.KeyMappingManager; 

public class SelectNoteScreen extends Screen {

    private static final Text TITLE = Text.translatable(Main.MOD_ID + ".screen.select_note.title");
    private static final Text INSTRUMENT_TEXT = Text.translatable(Main.MOD_ID + ".screen.select_note.instrument");
    private static final Text PITCH_TEXT = Text.translatable(Main.MOD_ID + ".screen.select_note.pitch");
    private static final Text DONE_BUTTON_TEXT = Text.translatable("gui.done");
    private static final Text CANCEL_BUTTON_TEXT = Text.translatable("gui.cancel");
    private static final Text PREVIEW_BUTTON_TEXT = Text.translatable(Main.MOD_ID + ".screen.select_note.preview");
    private final EditKeyMappingsScreen parent;
    private final InputUtil.Key keyToMap;

    private NoteBlockInstrument selectedInstrument = NoteBlockInstrument.HARP;
    private int selectedPitch = 12; 
    private CustomPitchSlider pitchSlider; 

    private ButtonWidget previewButton;

    public SelectNoteScreen(EditKeyMappingsScreen parent, InputUtil.Key keyToMap) {
        super(TITLE);
        this.parent = parent;
        this.keyToMap = keyToMap;
    }

    @Override
    protected void init() {
        super.init();

        int centerX = this.width / 2;
        int startY = this.height / 2 - 50;
        int widgetWidth = 200;
        int widgetHeight = 20;
        int margin = 5;
        CyclingButtonWidget<NoteBlockInstrument> instrumentButton = CyclingButtonWidget.builder((NoteBlockInstrument instrument) -> Text.translatable("block.minecraft.note_block.instrument." + instrument.asString()))
                .values(NoteBlockInstrument.values())
                .initially(selectedInstrument)
                
                .build(centerX - widgetWidth / 2, startY, widgetWidth, widgetHeight, INSTRUMENT_TEXT, (button, instrument) -> {
                    this.selectedInstrument = instrument;
                    updatePitchSliderDisplay(); 
                    updatePreviewButton();
                });
        addDrawableChild(instrumentButton);
        pitchSlider = new CustomPitchSlider(centerX - widgetWidth / 2, startY + widgetHeight + margin, widgetWidth, widgetHeight, PITCH_TEXT, (selectedPitch / 24.0));
        updatePitchSliderDisplay();
        addDrawableChild(pitchSlider);
        previewButton = ButtonWidget.builder(PREVIEW_BUTTON_TEXT, button -> {
            playPreviewNote();
        }).dimensions(centerX - widgetWidth / 2, startY + (widgetHeight + margin) * 2, widgetWidth, widgetHeight).build();
        addDrawableChild(previewButton);
        int buttonWidth = 100;
        int buttonY = this.height - 30;
        int doneButtonX = centerX - buttonWidth - margin;
        addDrawableChild(ButtonWidget.builder(DONE_BUTTON_TEXT, button -> {
            Note selectedNote = new Note(selectedInstrument, (byte) selectedPitch);
            parent.addNewMapping(keyToMap, selectedNote); 
            this.client.setScreen(this.parent);
            parent.stopWaitingForKeyPress(); 
        }).dimensions(doneButtonX, buttonY, buttonWidth, widgetHeight).build());
        int cancelButtonX = centerX + margin;
        addDrawableChild(ButtonWidget.builder(CANCEL_BUTTON_TEXT, button -> {
            this.client.setScreen(this.parent); 
            parent.stopWaitingForKeyPress(); 
        }).dimensions(cancelButtonX, buttonY, buttonWidth, widgetHeight).build());
    }
    private class CustomPitchSlider extends SliderWidget {
        public CustomPitchSlider(int x, int y, int width, int height, Text text, double value) {
            super(x, y, width, height, text, value);
        }

        @Override
        protected void updateMessage() {
            
            this.setMessage(PITCH_TEXT.copy().append(": " + KeyMappingManager.getNoteDisplayName(new Note(selectedInstrument, (byte) selectedPitch))));
        }

        @Override
        protected void applyValue() {
            selectedPitch = (int) Math.round(MathHelper.lerp(this.value, 0.0, 24.0));
            updateMessage(); 
            updatePreviewButton();
        }
        public void forceUpdateDisplay() {
            this.updateMessage(); 
        }
    }
    private void updatePitchSliderDisplay() {
        if (pitchSlider != null) {
            
            pitchSlider.forceUpdateDisplay();
        }
    }

    private void updatePreviewButton() {
        
    }

    private void playPreviewNote() {
        MinecraftClient client = MinecraftClient.getInstance();
        if (client.world != null && client.gameRenderer != null) {
            Vec3d pos = client.gameRenderer.getCamera().getPos();
            try {
                Note note = new Note(selectedInstrument, (byte) selectedPitch);
                
                if (note.instrument().canBePitched()) {
                    Identifier soundId = note.instrument().getSound().value().id();
                    float pitchMultiplier = (float) Math.pow(2.0, (note.note() - 12) / 12.0);
                    client.world.playSound(pos.x, pos.y, pos.z, SoundEvent.of(soundId), SoundCategory.RECORDS, 3.0f, pitchMultiplier, false);
                } else {
                    
                    Identifier soundId = note.instrument().getSound().value().id();
                    client.world.playSound(pos.x, pos.y, pos.z, SoundEvent.of(soundId), SoundCategory.RECORDS, 3.0f, 1.0f, false);
                }

            } catch (Exception e) {
                Main.LOGGER.error("无法播放预览声音。", e);
            }
        }
    }
    @Override
    public void render(DrawContext context, int mouseX, int mouseY, float delta) {

        super.render(context, mouseX, mouseY, delta);

        context.drawCenteredTextWithShadow(textRenderer, TITLE, this.width / 2, 10, 0xFFFFFF);
        
        context.drawCenteredTextWithShadow(textRenderer, Text.translatable(Main.MOD_ID + ".screen.select_note.mapping_key", Text.translatable(keyToMap.getTranslationKey())), this.width / 2, 30, 0xFFFFFF);
    }

    @Override
    public boolean shouldPause() {
        return true;
    }

    @Override
    public boolean keyPressed(int keyCode, int scanCode, int modifiers) {
        
        if (keyCode == GLFW.GLFW_KEY_ESCAPE) {
            this.client.setScreen(this.parent); 
            parent.stopWaitingForKeyPress(); 
            return true;
        }
        
        if (keyCode == GLFW.GLFW_KEY_ENTER || keyCode == GLFW.GLFW_KEY_KP_ENTER) {
            Note selectedNote = new Note(selectedInstrument, (byte) selectedPitch);
            parent.addNewMapping(keyToMap, selectedNote); 
            this.client.setScreen(this.parent); 
            parent.stopWaitingForKeyPress(); 
            return true;
        }
        return super.keyPressed(keyCode, scanCode, modifiers);
    }
}
