package semmiedev.disc_jockey_revive.gui.widget;

import net.minecraft.client.gui.DrawContext;
import net.minecraft.client.gui.screen.narration.NarrationMessageBuilder;
import net.minecraft.client.gui.widget.ClickableWidget;
import net.minecraft.text.Text;
import net.minecraft.util.math.MathHelper;

public class ProgressBarWidget extends ClickableWidget {
    private double progress;
    private final int minValue;
    private final int maxValue;
    private boolean dragging;

    public ProgressBarWidget(int x, int y, int width, int height, Text message, int minValue, int maxValue) {
        super(x, y, width, height, message);
        this.minValue = minValue;
        this.maxValue = maxValue;
        this.progress = minValue;
    }

    @Override
    public void renderWidget(DrawContext context, int mouseX, int mouseY, float delta) {
        // 背景
        context.fill(getX(), getY(), getX() + width, getY() + height, 0xFF555555);

        // 进度条
        int progressWidth = (int)(width * (progress - minValue) / (maxValue - minValue));
        context.fill(getX(), getY(), getX() + progressWidth, getY() + height, 0xFF00FF00);

        // 滑块
        int sliderX = getX() + progressWidth - 3;
        context.fill(sliderX, getY() - 2, sliderX + 6, getY() + height + 2, 0xFFFFFFFF);
    }

    @Override
    public boolean mouseClicked(double mouseX, double mouseY, int button) {
        if (button == 0 && isMouseOver(mouseX, mouseY)) {
            setProgressFromMouse(mouseX);
            dragging = true;
            return true;
        }
        return false;
    }

    @Override
    public boolean mouseDragged(double mouseX, double mouseY, int button, double deltaX, double deltaY) {
        if (dragging) {
            setProgressFromMouse(mouseX);
            return true;
        }
        return false;
    }

    @Override
    public boolean mouseReleased(double mouseX, double mouseY, int button) {
        if (button == 0) {
            dragging = false;
            return true;
        }
        return false;
    }

    private void setProgressFromMouse(double mouseX) {
        double relativeX = MathHelper.clamp(mouseX - getX(), 0, width);
        double newProgress = (relativeX / width) * (maxValue - minValue) + minValue;
        setProgress(newProgress);
    }

    public void setProgress(double progress) {
        this.progress = MathHelper.clamp(progress, minValue, maxValue);
        onProgressChanged(this.progress);
    }

    public double getProgress() {
        return progress;
    }

    public boolean isDragging() {
        return this.dragging;
    }

    protected void onProgressChanged(double progress) {

    }

    @Override
    protected void appendClickableNarrations(NarrationMessageBuilder builder) {
        appendDefaultNarrations(builder);
    }
}
